#ifndef __GENERAL_H
#define __GENERAL_H 1


#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <limits.h>
#include <sys/stat.h>
//#include <sys/types.h>

#include "types.h"

#ifdef ASM_X86
# include "asm_x86.h"
#endif

#if USE_BYTESWAP_H
# include <byteswap.h>
# define BSWAP_16(B) 	(bswap_16 (B))
# define BSWAP_32(B)	(bswap_32 (B))
# define BSWAP_64(B)	(bswap_64 (B))
#else
# define BSWAP_16(B)\
	({\
			__u16 _B = (B);\
			(((__u16) ((_B) << 8) | ((_B) >> 8)));\
	})
# define BSWAP_32(B)\
	({\
			__u32 _B = (B);\
			(((__u32) (_B << 24) | ((_B << 8) & 0xff0000) | ((_B >> 8) & 0xff00) | (_B >> 24)));\
	})
# define BSWAP_64(B)\
	({\
			__u64 _B = (B);\
			__u32 *_B32 = (__u32 *) &_B;\
			(((__u64) BSWAP32 (_B32[0])) << 32) | BSWAP32 (_B32[1]);\
	})
#endif

#ifndef CARRY3
# define CARRY3(a,b,c)				(((a) + (b) < a) || ((a) + (b) + (c) < c))
#endif

#ifndef CALC_OVERFLOW
static inline int CALC_OVERFLOW (__s32 a, __s32 b)
{
	return ((a > 0) && (b > 0) && (INT_MAX - a < b)) || ((a < 0) && (b < 0) && (-INT_MAX - a > b));
}
#endif

#ifndef CALC_OVERFLOW3
# define CALC_OVERFLOW3(a,b,c) 		(CALC_OVERFLOW (a,b) || CALC_OVERFLOW (a+b,c))
#endif

#ifndef CALC_UNDERFLOW
static inline int CALC_UNDERFLOW (__s32 a, __s32 b)
{
	return ((a > 0) && (b < 0) && (INT_MAX + b < a)) || ((a < 0) && (b > 0) && (-INT_MAX + b > a));
}
#endif

#ifndef CALC_UNDERFLOW3
# define CALC_UNDERFLOW3(a,b,c)		(CALC_UNDERFLOW (a,b) || CALC_UNDERFLOW (a+b,c))
#endif

#ifndef ROTL
# define ROTL(X,n)\
	({\
			__u32 _X = (X);\
			__u32 _n = (n);\
			((_X << _n) | (_X >> (32 - _n)));\
	})
#endif


#ifndef FALSE
# define FALSE 0
#endif

#ifndef TRUE
# define TRUE !FALSE
#endif

#include <SDL/SDL_byteorder.h>
#if SDL_BYTEORDER == SDL_LIL_ENDIAN
# define LIL_ENDIAN 1
#endif


#define MIN(a,b)					(((a) < (b)) ? (a) : (b))
#define CLAMP(X,min,max)	(((X) > max) ? max : (((X) < min) ? min : (X)))
#define EPSI							0.0001
#define ABS(X)						(((X) > 0) ? X : -X)
#define FLOAT_EQ(a,b)			(ABS ((a) - (b)) < EPSI)

#define CRC_POLY		0xA0000001


unsigned int file_exists (char *filename);
unsigned int file_size (char *filename);

void log_open (char *filename);
void log_printf (char *fmt, ...);
void log_close (void);

char *get_filename (char *name, const char *path);
char *get_path (char *path, const char *name);
char *get_extension (char *ext, const char *filename);
char *kill_extension (char *filename);

char *get_home_dir (void);

__u32 round_up (__u32 a, __u32 b);
inline int is_power_of_two (__u32 a);

__u32 crc_setup (unsigned int bits);
__u32 crc_iterate (__u32 crc, __u8 d);

void create_dir (char *name);
void create_dir_tree (char *parent);

int path_writeable (char *path);


#endif // __GENERAL_H
