/*
 *  gcube
 *  Copyright (c) 2004 monk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *
 *  SI - serial interface (0xcc006400, 0x40)
 *      
 *         
 */

#include "hw_si.h"


__u8 rsi[RSI_SIZE];
int data_available[4] = {0};


int controller_type[4] =
{
	SIDEV_GC_CONTROLLER,
	0,
	0,
	0,
};

__u32 si_r32_direct (__u32 addr)
{
	DEBUG (EVENT_LOG_SI, "..si: read  [%.4x] (%.8x)", addr & 0xffff, RSI32 (addr));
	return RSI32 (addr);
}


void si_w32_direct (__u32 addr, __u32 data)
{
	DEBUG (EVENT_LOG_SI, "..si: write [%.4x] (%.8x)", addr & 0xffff, data);
	RSI32 (addr) = data;
}


__u32 si_r32_input (__u32 addr)
{
	int rdst[] = 
	{
		SI_SR_RDST0, SI_SR_RDST1, SI_SR_RDST2, SI_SR_RDST3,
	};
	int port = ((addr - SI_CH0_INH) & 0xff) / 0x0c;
	

	data_available[port] = FALSE;

	RSI32 (SI_SR) &= ~rdst[port];
	if (!(RSI32 (SI_SR) & (SI_SR_RDST0 | SI_SR_RDST1 | SI_SR_RDST2 | SI_SR_RDST3)))
		RSI32 (SI_COMMCSR) &= ~SI_COMMCSR_RDSTINT;

	return RSI32 (addr);
}


void si_w32_commcsr (__u32 addr, __u32 data)
{
	DEBUG (EVENT_LOG_SI, "..si: write [%.4x] (%.8x) CSR", addr & 0xffff, data);

	// acknowlegde interrupt
	if (data & SI_COMMCSR_TCINT)
		RSI32 (addr) &= ~SI_COMMCSR_TCINT;

	if (data & SI_COMMCSR_RDSTINTMSK)
		RSI32 (addr) |= SI_COMMCSR_RDSTINTMSK;

	if (data & SI_COMMCSR_TSTART)
	{
		// do transfer
		DEBUG (EVENT_LOG_SI, "..... COMMAND %.2x", RSI8 (SI_IO_BUFF));
		switch (RSI8 (SI_IO_BUFF))
		{
			// get device type and status
			case 0x00:
				RSI32 (SI_IO_BUFF) = controller_type[(data >> 1) & 3];
				break;
			
			// read origins
			case 0x41:
				RSI8 (SI_IO_BUFF + 1) = 0;

				RSI8 (SI_IO_BUFF + 2) = 0x80;
				RSI8 (SI_IO_BUFF + 3) = 0x80;
				RSI8 (SI_IO_BUFF + 4) = 0x80;
				RSI8 (SI_IO_BUFF + 5) = 0x80;

				RSI8 (SI_IO_BUFF + 6) = 0x1f;
				RSI8 (SI_IO_BUFF + 7) = 0x1f;
				//??
				RSI8 (SI_IO_BUFF + 8) = 0x1f;
				RSI8 (SI_IO_BUFF + 9) = 0x1f;
				break;
		}
		
		// complete transfer
		RSI32 (addr) &= ~SI_COMMCSR_TSTART;

		// transfer complete interrupt
		RSI32 (addr) |= SI_COMMCSR_TCINT;
		if (data & SI_COMMCSR_TCINTMSK)
		{
			RSI32 (addr) |= SI_COMMCSR_TCINTMSK;
			pi_interrupt (INTERRUPT_SI);
		}
	}
}


void si_w32_sipoll (__u32 addr, __u32 data)
{
	DEBUG (EVENT_LOG_SI, "..si: write [%.4x] (%.8x) SIPOLL", addr & 0xffff, data);

	if (data & SI_POLL_EN0)
		RSI32 (SI_SR) |= SI_SR_RDST0;

	if (data & SI_POLL_EN1)
		RSI32 (SI_SR) |= SI_SR_RDST1;

	if (data & SI_POLL_EN2)
		RSI32 (SI_SR) |= SI_SR_RDST2;

	if (data & SI_POLL_EN3)
		RSI32 (SI_SR) |= SI_SR_RDST3;

	RSI32 (addr) = data;
}


void si_w32_sr (__u32 addr, __u32 data)
{
	data &= ~(SI_SR_WRST0 | SI_SR_WRST1 | SI_SR_WRST2 | SI_SR_WRST3 | SI_SR_WR);

	DEBUG (EVENT_LOG_SI, "..si: write [%.4x] (%.8x) SR", addr & 0xffff, data);
	RSI32 (addr) = data;
}


void pad_set_action (int n, __u16 buttons, __s8 xstick, __s8 ystick,
										 __s8 cxstick, __s8 cystick, __s8 trigl, __s8 trigr)
{
	RSI32 (SI_CHN_INH (n)) = SI_CHH (buttons, xstick, ystick);
	RSI32 (SI_CHN_INL (n)) = SI_CHL (cxstick, cystick, trigl, trigr);

	RSI32 (SI_CHN_INH (n)) |= 0x00800000; //?
	data_available[n] = TRUE;
}


// for keyboard mapping
void pad_action (int n, int pressed, int action)
{
	if (action < ACTION_AXIS)
		pad_action_buttons (n, pressed, action);
	else
	{
		action -= ACTION_AXIS;
		// axis0-, axis0+, ...
		if (pressed)
		{
			if (action & 2)
				pad_action_axis (n, action / 2, (action & 1) ? 127 : -128);
			else
				pad_action_axis (n, action / 2, (action & 1) ? -128 : 127);
		}
		else
			pad_action_axis (n, action / 2, 0);
	}

	data_available[n] = TRUE;
}


void pad_action_buttons (int n, int pressed, __u16 buttons)
{
	if (pressed)
		RSI32 (SI_CHN_INH (n)) |=  (buttons << 16);
	else
		RSI32 (SI_CHN_INH (n)) &= ~(buttons << 16);

	data_available[n] = TRUE;
}


void pad_action_buttons_masked (int n, int pressed, __u16 buttons, int mask)
{
	RSI32 (SI_CHN_INH (n)) &= ~(mask << 16);
	pad_action_buttons (n, pressed, buttons);
}


void pad_action_analogs (int n, __s8 x, __s8 y)
{
	RSI32 (SI_CHN_INH (n)) = (RSI32 (SI_CHN_INH (n)) & 0xffff0000) |
													((__u8) x << 8) | ((__u8) y);

	RSI32 (SI_CHN_INH (n)) |= 0x00800000; //?

	data_available[n] = TRUE;
}


void pad_action_canalogs (int n, __s8 x, __s8 y)
{
	RSI32 (SI_CHN_INL (n)) = (RSI32 (SI_CHN_INL (n)) & 0x0000ffff) |
													((__u8) x << 24) | ((__u8) y << 16);

	RSI32 (SI_CHN_INH (n)) |= 0x00800000; //?

	data_available[n] = TRUE;
}


void pad_action_triggers (int n, __s8 l, __s8 r)
{
	RSI32 (SI_CHN_INH (n)) = (RSI32 (SI_CHN_INH (n)) & 0xffff0000) |
													((__u8) l << 8) | ((__u8) r);

	data_available[n] = TRUE;
}


void pad_action_axis (int n, int axis, __s8 value)
{
	switch (axis)
	{
		// analog y
		case 0:
			RSI32 (SI_CHN_INH (n)) = (RSI32 (SI_CHN_INH (n)) & 0xffffff00) |
															((__u8) (0x80 + value));
			break;

		// analog x
		case 1:
			RSI32 (SI_CHN_INH (n)) = (RSI32 (SI_CHN_INH (n)) & 0xffff00ff) |
															((__u8) (0x80 + value) << 8);
			break;

		// canalog y
		case 2:
			RSI32 (SI_CHN_INL (n)) = (RSI32 (SI_CHN_INL (n)) & 0xff00ffff) |
															((__u8) (0x80 + value) << 16);
			break;
		
		// canalog x
		case 3:
			RSI32 (SI_CHN_INL (n)) = (RSI32 (SI_CHN_INL (n)) & 0x00ffffff) |
															((__u8) (0x80 + value) << 24);
			break;

		default:
			break;
	}

	RSI32 (SI_CHN_INH (n)) |= 0x00800000; //?

	data_available[n] = TRUE;
}


void si_poll (void)
{
	if (RSI32 (SI_POLL) & SI_POLL_EN0)
	{
		if (data_available[0])
		{
		RSI32 (SI_SR) |= SI_SR_RDST0;
			RSI32 (SI_COMMCSR) |= SI_COMMCSR_RDSTINT;
			if (RSI32 (SI_COMMCSR) & SI_COMMCSR_RDSTINTMSK)
				pi_interrupt_ex (INTERRUPT_SI);
		}
	}

	if (RSI32 (SI_POLL) & SI_POLL_EN1)
	{
		if (data_available[1])
		{
		RSI32 (SI_SR) |= SI_SR_RDST1;
			RSI32 (SI_COMMCSR) |= SI_COMMCSR_RDSTINT;
			if (RSI32 (SI_COMMCSR) & SI_COMMCSR_RDSTINTMSK)
				pi_interrupt_ex (INTERRUPT_SI);
		}
	}

	if (RSI32 (SI_POLL) & SI_POLL_EN2)
	{
		if (data_available[2])
		{
		RSI32 (SI_SR) |= SI_SR_RDST2;
			RSI32 (SI_COMMCSR) |= SI_COMMCSR_RDSTINT;
			if (RSI32 (SI_COMMCSR) & SI_COMMCSR_RDSTINTMSK)
				pi_interrupt_ex (INTERRUPT_SI);
		}
	}

	if (RSI32 (SI_POLL) & SI_POLL_EN3)
	{
		if (data_available[3])
		{
		RSI32 (SI_SR) |= SI_SR_RDST3;
			RSI32 (SI_COMMCSR) |= SI_COMMCSR_RDSTINT;
			if (RSI32 (SI_COMMCSR) & SI_COMMCSR_RDSTINTMSK)
				pi_interrupt_ex (INTERRUPT_SI);
		}
	}
}


void si_reinit (void)
{
}


void si_init (void)
{
	int i;


	memset (rsi, 0, sizeof (rsi));

	// initialize pads
	for (i = 0; i < 4; i++)
	{
		// pad command
		RSI32 (SI_CHN_CMD (i)) = 0x00400300;		// ?

		// center analogs
		RSI32 (SI_CHN_INH (i)) = SI_CHH (0, 0, 0);
		RSI32 (SI_CHN_INL (i)) = SI_CHL (0, 0, 0, 0);
		mem_hwr_hook (32, SI_CHN_CMD (i), si_r32_direct);
		mem_hww_hook (32, SI_CHN_CMD (i), si_w32_direct);

		// clear rdst on read
		mem_hwr_hook (32, SI_CHN_INH (i), si_r32_input);
		mem_hww_hook (32, SI_CHN_INH (i), si_w32_direct);
		mem_hwr_hook (32, SI_CHN_INL (i), si_r32_input);
		mem_hww_hook (32, SI_CHN_INL (i), si_w32_direct);
	}

	// sipoll
	mem_hwr_hook (32, SI_POLL, si_r32_direct);	
	mem_hww_hook (32, SI_POLL, si_w32_sipoll);

	// sicommcsr
	mem_hwr_hook (32, SI_COMMCSR, si_r32_direct);	
	mem_hww_hook (32, SI_COMMCSR, si_w32_commcsr);	

	// sisr - si status register
	mem_hwr_hook (32, SI_SR, si_r32_direct);	
	mem_hww_hook (32, SI_SR, si_w32_sr);	

	// si io buffer
	for (i = 0; i < 128; i += 4)
	{
		mem_hww_hook (32, SI_IO_BUFF + i, si_w32_direct);
		mem_hwr_hook (32, SI_IO_BUFF + i, si_r32_direct);
	}
}
